/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *     I use LibSVM, so please make sure you correctly include the class path for libsvm    
 *    
 *    
 *    SubspaceSVM.java
 *    Copyright (C) 2008 Tzu-Cheng Chuang
 *
 */
package weka.classifiers.TzuCheng;

import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

import weka.classifiers.trees.j48.*;
import weka.classifiers.functions.LibSVM;
import weka.classifiers.*;
import weka.core.*;


//import weka.core.Instances;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.*;
import weka.filters.Filter;

public class SubspaceSVM 
	extends Classifier 
	implements TechnicalInformationHandler { 
	
	static final long serialVersionUID = -6585883627778691736L;
	protected BinC45Split modSplit[] ;
	protected LibSVM svm1[];
	protected LibSVM svm2[];
	protected int svm_structure[];
	protected int pick_topn;
	protected boolean use_this_split[];
	//%Pick the top 1/8 (->16)high training accuracy for consensus
	//%Pick the top 1/4 (->8)high training accuracy for consensus

	
	/** LibSVM Model */
	  protected Object m_Model;
	  
	  /** for normalizing the data */
	  protected Filter m_Filter = null;
	  
	  /** normalize input data */
	  protected boolean m_Normalize = false;
	  
	  /** SVM type C-SVC (classification) */
	  public static final int SVMTYPE_C_SVC = 0;
	  /** SVM type nu-SVC (classification) */
	  public static final int SVMTYPE_NU_SVC = 1;
	  /** SVM type one-class SVM (classification) */
	  public static final int SVMTYPE_ONE_CLASS_SVM = 2;
	  /** SVM type epsilon-SVR (regression) */
	  public static final int SVMTYPE_EPSILON_SVR = 3;
	  /** SVM type nu-SVR (regression) */
	  public static final int SVMTYPE_NU_SVR = 4;
	  /** SVM types */
	  public static final Tag[] TAGS_SVMTYPE = {
	    new Tag(SVMTYPE_C_SVC, "C-SVC (classification)"),
	    new Tag(SVMTYPE_NU_SVC, "nu-SVC (classification)"),
	    new Tag(SVMTYPE_ONE_CLASS_SVM, "one-class SVM (classification)"),
	    new Tag(SVMTYPE_EPSILON_SVR, "epsilon-SVR (regression)"),
	    new Tag(SVMTYPE_NU_SVR, "nu-SVR (regression)")
	  };
	  
	  /** the SVM type */
	  protected int m_SVMType = SVMTYPE_C_SVC;
	  
	  /** kernel type linear: u'*v */
	  public static final int KERNELTYPE_LINEAR = 0;
	  /** kernel type polynomial: (gamma*u'*v + coef0)^degree */
	  public static final int KERNELTYPE_POLYNOMIAL = 1;
	  /** kernel type radial basis function: exp(-gamma*|u-v|^2) */
	  public static final int KERNELTYPE_RBF = 2;
	  /** kernel type sigmoid: tanh(gamma*u'*v + coef0) */
	  public static final int KERNELTYPE_SIGMOID = 3;
	  /** the different kernel types */
	  public static final Tag[] TAGS_KERNELTYPE = {
	    new Tag(KERNELTYPE_LINEAR, "linear: u'*v"),
	    new Tag(KERNELTYPE_POLYNOMIAL, "polynomial: (gamma*u'*v + coef0)^degree"),
	    new Tag(KERNELTYPE_RBF, "radial basis function: exp(-gamma*|u-v|^2)"),
	    new Tag(KERNELTYPE_SIGMOID, "sigmoid: tanh(gamma*u'*v + coef0)")
	  };
	  
	  /** the kernel type */
	  protected int m_KernelType = KERNELTYPE_RBF;
	  
	  /** for poly - in older versions of libsvm declared as a double.
	   * At least since 2.82 it is an int. */
	  protected int m_Degree = 3;
	  
	  /** for poly/rbf/sigmoid */
	  protected double m_Gamma = 0;
	  
	  /** for poly/rbf/sigmoid (the actual gamma) */
	  protected double m_GammaActual = 0;
	  
	  /** for poly/sigmoid */
	  protected double m_Coef0 = 0;
	  
	  /** in MB */
	  protected double m_CacheSize = 40;
	  
	  /** stopping criteria */
	  protected double m_eps = 1e-3;
	  
	  /** cost, for C_SVC, EPSILON_SVR and NU_SVR */
	  protected double m_Cost = 1;
	  
	  /** for C_SVC */
	  protected int[] m_WeightLabel = new int[0];
	  
	  /** for C_SVC */
	  protected double[] m_Weight = new double[0];
	  
	  /** for NU_SVC, ONE_CLASS, and NU_SVR */
	  protected double m_nu = 0.5;
	  
	  /** loss, for EPSILON_SVR */
	  protected double m_Loss = 0.1;
	  
	  /** use the shrinking heuristics */
	  protected boolean m_Shrinking = true;	
	  
	  /** whether to generate probability estimates instead of +1/-1 in case of 
	   * classification problems */
	  protected boolean m_ProbabilityEstimates = false;
	
	
	
	//svm_structure=0, svm1 and svm2 both exist
	//svm_structure=1, only svm1 exists
	//svm_structure=2, only svm2 exists
	
	public String globalInfo() {
	    return  "Implements Tzu-Cheng Chuang's Subspace SVM"
	      + "For more information on tthis algorithm, see\n\n"
	      + getTechnicalInformation().toString();
	  }
	/**
	   * Returns an instance of a TechnicalInformation object, containing 
	   * detailed information about the technical background of this class,
	   * e.g., paper reference or book this class is based on.
	   * 
	   * @return the technical information about this class
	   */
	public TechnicalInformation getTechnicalInformation() {
	    TechnicalInformation 	result;
	    
	    result = new TechnicalInformation(Type.INCOLLECTION);
	    result.setValue(Field.AUTHOR, "Tzu-Cheng Chuang");
	    result.setValue(Field.YEAR, "2008");
	    result.setValue(Field.TITLE, "Machines ");
	    result.setValue(Field.BOOKTITLE, "ANNIE");
	    result.setValue(Field.EDITOR, "Someone");
	    result.setValue(Field.PUBLISHER, "ACM Press");
	    result.setValue(Field.URL, "http://smo.html");
	    result.setValue(Field.PDF, "http://smo-book.pdf");
	    result.setValue(Field.PS, "http://smo-book.ps.gz");
	            
	    return result;
	  }
	/**
	   * Returns default capabilities of the classifier.
	   *
	   * @return      the capabilities of this classifier
	   */
	  public Capabilities getCapabilities() {
	    Capabilities result = super.getCapabilities();

	    // attributes
	    result.enableAllAttributeDependencies();
	    
	    result.enable(Capability.NOMINAL_ATTRIBUTES);
	    result.enable(Capability.NUMERIC_ATTRIBUTES);
	    result.enable(Capability.DATE_ATTRIBUTES);
	    
	    // class
	    result.disableAllClasses();
	    result.disableAllClassDependencies();
	    result.enable(Capability.NOMINAL_CLASS);
	    result.enable(Capability.MISSING_CLASS_VALUES);
	    
	    return result;
	  }
	  
	/**
	   * Method for building the classifier. Implements a one-against-one
	   * wrapper for multi-class problems.
	   *
	   * @param insts the set of training instances
	   * @throws Exception if the classifier can't be built successfully
	   */
	public void buildClassifier(Instances data) throws Exception {
	  int m_minNoObj=2;  
	  double sumOfWeights = data.sumOfWeights();
	  modSplit = new BinC45Split[data.numAttributes()];
	  for (int i_att=0;i_att<data.numAttributes();i_att++){
	    modSplit[i_att]=new BinC45Split(i_att,m_minNoObj,sumOfWeights);
	    modSplit[i_att].buildClassifier(data);
	  }
	  svm1 = new LibSVM[data.numAttributes()];
	  svm2 = new LibSVM[data.numAttributes()];
	  for (int i_att=0;i_att<data.numAttributes();i_att++){
	    String svm_options_copy[]= getOptions().clone();
	    svm1[i_att]=new LibSVM();
	    svm1[i_att].setOptions(svm_options_copy);
	    //svm1[i_att].setOptions(svm_options);
	    
	    String svm_options_copy2[]= getOptions().clone();
	    svm2[i_att]=new LibSVM();
	    svm2[i_att].setOptions(svm_options_copy2);
		  
	  }
	  
	  svm_structure = new int[data.numAttributes()];
	  
	  for (int i_att=0;i_att<data.numAttributes();i_att++){
	    int m_numSubsets=2;
	    Instances [] instances = new Instances [m_numSubsets];
	    double [] weights;
	    double newWeight;
	    Instance instance;
	    int subset, i, j;

	    for (j=0;j<m_numSubsets;j++)
	      instances[j] = new Instances((Instances)data,data.numInstances());
	    
	    for (i = 0; i < data.numInstances(); i++) {
	      instance = ((Instances) data).instance(i);
	      weights = modSplit[i_att].weights(instance);
	      subset = modSplit[i_att].whichSubset(instance);
	      if (subset > -1)
	    	  instances[subset].add(instance);
	      else
	    	  for (j = 0; j < m_numSubsets; j++)
	    		  if (Utils.gr(weights[j],0)) {
	    			  newWeight = weights[j]*instance.weight();
	    			  instances[j].add(instance);
	    			  instances[j].lastInstance().setWeight(newWeight);
	    		  }
	    	}
	    for (j = 0; j < m_numSubsets; j++)
	      instances[j].compactify();
	    
	    if (0!=instances[0].numInstances()) 
	      svm1[i_att].buildClassifier(instances[0]);
	    else
	      svm_structure[i_att]=2;
	    
	    if (0!=instances[1].numInstances())
	      svm2[i_att].buildClassifier(instances[1]);  
	    else
	      svm_structure[i_att]=1;
	  }
	  double info_split[] = new double[data.numAttributes()-1];
	  
	  for(int i=0;i<data.numAttributes()-1;i++)
	    info_split[i]=modSplit[i].infoGain();
	  info_split[data.numAttributes()-2]=-5000;
	  
	  
	  int order_idx[]=Utils.sort(info_split);
	  
	  int cap_range= (int)Math.ceil(((double)data.numAttributes())/8);
	  pick_topn=cap_range*2+1;
	  
	  use_this_split=new boolean[data.numAttributes()];
	  for (int i=0;i<pick_topn;i++){
	    //System.out.println("hihiSubspaceSVM"+data.numAttributes());
	    use_this_split[order_idx[data.numAttributes()-i-2]]=true;
	  }
	  //System.out.println("hihiSubspaceSVM"+data.numAttributes());
	}
	/**
	   * Returns an enumeration describing the available options.
	   * 
	   * @return an enumeration of all the available options.
	   */
	  public Enumeration listOptions() {
	    Vector      result;
	    
	    result = new Vector();
	    
	    result.addElement(
	        new Option(
	            "\tSet type of SVM (default: 0)\n"
	            + "\t\t 0 = C-SVC\n" 
	            + "\t\t 1 = nu-SVC\n"
	            + "\t\t 2 = one-class SVM\n" 
	            + "\t\t 3 = epsilon-SVR\n"
	            + "\t\t 4 = nu-SVR", 
	            "S", 1, "-S <int>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet type of kernel function (default: 2)\n"
	            + "\t\t 0 = linear: u'*v\n"
	            + "\t\t 1 = polynomial: (gamma*u'*v + coef0)^degree\n"
	            + "\t\t 2 = radial basis function: exp(-gamma*|u-v|^2)\n"
	            + "\t\t 3 = sigmoid: tanh(gamma*u'*v + coef0)",
	            "K", 1, "-K <int>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet degree in kernel function (default: 3)", 
	            "D", 1, "-D <int>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet gamma in kernel function (default: 1/k)", 
	            "G", 1, "-G <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet coef0 in kernel function (default: 0)", 
	            "R", 1, "-R <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet the parameter C of C-SVC, epsilon-SVR, and nu-SVR\n"
	            + "\t (default: 1)",
	            "C", 1, "-C <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet the parameter nu of nu-SVC, one-class SVM, and nu-SVR\n"
	            + "\t (default: 0.5)",
	            "N", 1, "-N <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tTurns on normalization of input data (default: off)", 
	            "Z", 0, "-Z"));
	    
	    result.addElement(
	        new Option(
	            "\tSet the epsilon in loss function of epsilon-SVR (default: 0.1)",
	            "P", 1, "-P <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet cache memory size in MB (default: 40)", 
	            "M", 1, "-M <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tSet tolerance of termination criterion (default: 0.001)",
	            "E", 1, "-E <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tTurns the shrinking heuristics off (default: on)",
	            "H", 0, "-H"));
	    
	    result.addElement(
	        new Option(
	            "\tSet the parameters C of class i to weight[i]*C, for C-SVC\n" 
	            + "\t (default: 1)",
	            "W", 1, "-W <double>"));
	    
	    result.addElement(
	        new Option(
	            "\tTrains a SVC model instead of a SVR one (default: SVR)",
	            "B", 0, "-B"));

	    Enumeration en = super.listOptions();
	    while (en.hasMoreElements())
	      result.addElement(en.nextElement());
	    
	    return result.elements();
	  }
	  

	  /**
	   * Sets the classifier options <p/>
	   * 
	   <!-- options-start -->
	   * Valid options are: <p/>
	   * 
	   * <pre> -S &lt;int&gt;
	   *  Set type of SVM (default: 0)
	   *    0 = C-SVC
	   *    1 = nu-SVC
	   *    2 = one-class SVM
	   *    3 = epsilon-SVR
	   *    4 = nu-SVR</pre>
	   * 
	   * <pre> -K &lt;int&gt;
	   *  Set type of kernel function (default: 2)
	   *    0 = linear: u'*v
	   *    1 = polynomial: (gamma*u'*v + coef0)^degree
	   *    2 = radial basis function: exp(-gamma*|u-v|^2)
	   *    3 = sigmoid: tanh(gamma*u'*v + coef0)</pre>
	   * 
	   * <pre> -D &lt;int&gt;
	   *  Set degree in kernel function (default: 3)</pre>
	   * 
	   * <pre> -G &lt;double&gt;
	   *  Set gamma in kernel function (default: 1/k)</pre>
	   * 
	   * <pre> -R &lt;double&gt;
	   *  Set coef0 in kernel function (default: 0)</pre>
	   * 
	   * <pre> -C &lt;double&gt;
	   *  Set the parameter C of C-SVC, epsilon-SVR, and nu-SVR
	   *   (default: 1)</pre>
	   * 
	   * <pre> -N &lt;double&gt;
	   *  Set the parameter nu of nu-SVC, one-class SVM, and nu-SVR
	   *   (default: 0.5)</pre>
	   * 
	   * <pre> -Z
	   *  Turns on normalization of input data (default: off)</pre>
	   * 
	   * <pre> -P &lt;double&gt;
	   *  Set the epsilon in loss function of epsilon-SVR (default: 0.1)</pre>
	   * 
	   * <pre> -M &lt;double&gt;
	   *  Set cache memory size in MB (default: 40)</pre>
	   * 
	   * <pre> -E &lt;double&gt;
	   *  Set tolerance of termination criterion (default: 0.001)</pre>
	   * 
	   * <pre> -H
	   *  Turns the shrinking heuristics off (default: on)</pre>
	   * 
	   * <pre> -W &lt;double&gt;
	   *  Set the parameters C of class i to weight[i]*C, for C-SVC
	   *   (default: 1)</pre>
	   * 
	   * <pre> -B
	   *  Trains a SVC model instead of a SVR one (default: SVR)</pre>
	   * 
	   * <pre> -D
	   *  If set, classifier is run in debug mode and
	   *  may output additional info to the console</pre>
	   * 
	   <!-- options-end -->
	   *
	   * @param options     the options to parse
	   * @throws Exception  if parsing fails
	   */
	  public void setOptions(String[] options) throws Exception {
	    String      tmpStr;
	    
	    tmpStr = Utils.getOption('S', options);
	    if (tmpStr.length() != 0)
	      setSVMType(
	          new SelectedTag(Integer.parseInt(tmpStr), TAGS_SVMTYPE));
	    else
	      setSVMType(
	          new SelectedTag(SVMTYPE_C_SVC, TAGS_SVMTYPE));
	    
	    tmpStr = Utils.getOption('K', options);
	    if (tmpStr.length() != 0)
	      setKernelType(
	          new SelectedTag(Integer.parseInt(tmpStr), TAGS_KERNELTYPE));
	    else
	      setKernelType(
	          new SelectedTag(KERNELTYPE_RBF, TAGS_KERNELTYPE));
	    
	    tmpStr = Utils.getOption('D', options);
	    if (tmpStr.length() != 0)
	      setDegree(Integer.parseInt(tmpStr));
	    else
	      setDegree(3);
	    
	    tmpStr = Utils.getOption('G', options);
	    if (tmpStr.length() != 0)
	      setGamma(Double.parseDouble(tmpStr));
	    else
	      setGamma(0);
	    
	    tmpStr = Utils.getOption('R', options);
	    if (tmpStr.length() != 0)
	      setCoef0(Double.parseDouble(tmpStr));
	    else
	      setCoef0(0);
	    
	    tmpStr = Utils.getOption('N', options);
	    if (tmpStr.length() != 0)
	      setNu(Double.parseDouble(tmpStr));
	    else
	      setNu(0.5);
	    
	    tmpStr = Utils.getOption('M', options);
	    if (tmpStr.length() != 0)
	      setCacheSize(Double.parseDouble(tmpStr));
	    else
	      setCacheSize(40);
	    
	    tmpStr = Utils.getOption('C', options);
	    if (tmpStr.length() != 0)
	      setCost(Double.parseDouble(tmpStr));
	    else
	      setCost(1);
	    
	    tmpStr = Utils.getOption('E', options);
	    if (tmpStr.length() != 0)
	      setEps(Double.parseDouble(tmpStr));
	    else
	      setEps(1e-3);
	    
	    setNormalize(Utils.getFlag('Z', options));
	    
	    tmpStr = Utils.getOption('P', options);
	    if (tmpStr.length() != 0)
	      setLoss(Double.parseDouble(tmpStr));
	    else
	      setLoss(0.1);
	    
	    setShrinking(!Utils.getFlag('H', options));
	    
	    setWeights(Utils.getOption('W', options));
	    
	    setProbabilityEstimates(Utils.getFlag('B', options));
	  }
	  
	  /**
	   * Returns the current options
	   * 
	   * @return            the current setup
	   */
	  public String[] getOptions() {
	    Vector        result;
	    
	    result  = new Vector();
	    
	    result.add("-S");
	    result.add("" + m_SVMType);
	    
	    result.add("-K");
	    result.add("" + m_KernelType);
	    
	    result.add("-D");
	    result.add("" + getDegree());
	    
	    result.add("-G");
	    result.add("" + getGamma());
	    
	    result.add("-R");
	    result.add("" + getCoef0());
	    
	    result.add("-N");
	    result.add("" + getNu());
	    
	    result.add("-M");
	    result.add("" + getCacheSize());
	    
	    result.add("-C");
	    result.add("" + getCost());
	    
	    result.add("-E");
	    result.add("" + getEps());
	    
	    result.add("-P");
	    result.add("" + getLoss());
	    
	    if (!getShrinking())
	      result.add("-H");
	    
	    if (getNormalize())
	      result.add("-Z");
	    
	    if (getWeights().length() != 0) {
	      result.add("-W");
	      result.add("" + getWeights());
	    }
	    
	    if (getProbabilityEstimates())
	      result.add("-B");
	    
	    return (String[]) result.toArray(new String[result.size()]);
	  }
	  /**
	   * Sets type of SVM (default SVMTYPE_C_SVC)
	   * 
	   * @param value       the type of the SVM
	   */
	  public void setSVMType(SelectedTag value) {
	    if (value.getTags() == TAGS_SVMTYPE)
	      m_SVMType = value.getSelectedTag().getID();
	  }
	  
	  /**
	   * Gets type of SVM
	   * 
	   * @return            the type of the SVM
	   */
	  public SelectedTag getSVMType() {
	    return new SelectedTag(m_SVMType, TAGS_SVMTYPE);
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String SVMTypeTipText() {
	    return "The type of SVM to use.";
	  }
	  
	  /**
	   * Sets type of kernel function (default KERNELTYPE_RBF)
	   * 
	   * @param value       the kernel type
	   */
	  public void setKernelType(SelectedTag value) {
	    if (value.getTags() == TAGS_KERNELTYPE)
	      m_KernelType = value.getSelectedTag().getID();
	  }
	  
	  /**
	   * Gets type of kernel function
	   * 
	   * @return            the kernel type
	   */
	  public SelectedTag getKernelType() {
	    return new SelectedTag(m_KernelType, TAGS_KERNELTYPE);
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String kernelTypeTipText() {
	    return "The type of kernel to use";
	  }
	  
	  /**
	   * Sets the degree of the kernel
	   * 
	   * @param value       the degree of the kernel
	   */
	  public void setDegree(int value) {
	    m_Degree = value;
	  }
	  
	  /**
	   * Gets the degree of the kernel
	   * 
	   * @return            the degree of the kernel
	   */
	  public int getDegree() {
	    return m_Degree;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String degreeTipText() {
	    return "The degree of the kernel.";
	  }
	  
	  /**
	   * Sets gamma (default = 1/no of attributes)
	   * 
	   * @param value       the gamma value
	   */
	  public void setGamma(double value) {
	    m_Gamma = value;
	  }
	  
	  /**
	   * Gets gamma
	   * 
	   * @return            the current gamma
	   */
	  public double getGamma() {
	    return m_Gamma;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String gammaTipText() {
	    return "The gamma to use, if 0 then 1/max_index is used.";
	  }
	  
	  /**
	   * Sets coef (default 0)
	   * 
	   * @param value       the coef
	   */
	  public void setCoef0(double value) {
	    m_Coef0 = value;
	  }
	  
	  /**
	   * Gets coef
	   * 
	   * @return            the coef
	   */
	  public double getCoef0() {
	    return m_Coef0;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String coef0TipText() {
	    return "The coefficient to use.";
	  }
	  
	  /**
	   * Sets nu of nu-SVC, one-class SVM, and nu-SVR (default 0.5)
	   * 
	   * @param value       the new nu value
	   */
	  public void setNu(double value) {
	    m_nu = value;
	  }
	  
	  /**
	   * Gets nu of nu-SVC, one-class SVM, and nu-SVR (default 0.5)
	   * 
	   * @return            the current nu value
	   */
	  public double getNu() {
	    return m_nu;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String nuTipText() {
	    return "The value of nu for nu-SVC, one-class SVM and nu-SVR.";
	  }
	  
	  /**
	   * Sets cache memory size in MB (default 40)
	   * 
	   * @param value       the memory size in MB
	   */
	  public void setCacheSize(double value) {
	    m_CacheSize = value;
	  }
	  
	  /**
	   * Gets cache memory size in MB
	   * 
	   * @return            the memory size in MB
	   */
	  public double getCacheSize() {
	    return m_CacheSize;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String cacheSizeTipText() {
	    return "The cache size in MB.";
	  }
	  
	  /**
	   * Sets the parameter C of C-SVC, epsilon-SVR, and nu-SVR (default 1)
	   * 
	   * @param value       the cost value
	   */
	  public void setCost(double value) {
	    m_Cost = value;
	  }
	  
	  /**
	   * Sets the parameter C of C-SVC, epsilon-SVR, and nu-SVR
	   * 
	   * @return            the cost value
	   */
	  public double getCost() {
	    return m_Cost;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String costTipText() {
	    return "The cost parameter C for C-SVC, epsilon-SVR and nu-SVR.";
	  }
	  
	  /**
	   * Sets tolerance of termination criterion (default 0.001)
	   * 
	   * @param value       the tolerance
	   */
	  public void setEps(double value) {
	    m_eps = value;
	  }
	  
	  /**
	   * Gets tolerance of termination criterion
	   * 
	   * @return            the current tolerance
	   */
	  public double getEps() {
	    return m_eps;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String epsTipText() {
	    return "The tolerance of the termination criterion.";
	  }
	  
	  /**
	   * Sets the epsilon in loss function of epsilon-SVR (default 0.1)
	   * 
	   * @param value       the loss epsilon
	   */
	  public void setLoss(double value) {
	    m_Loss = value;
	  }
	  
	  /**
	   * Gets the epsilon in loss function of epsilon-SVR
	   * 
	   * @return            the loss epsilon
	   */
	  public double getLoss() {
	    return m_Loss;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String lossTipText() {
	    return "The epsilon for the loss function in epsilon-SVR.";
	  }
	  
	  /**
	   * whether to use the shrinking heuristics
	   * 
	   * @param value       true uses shrinking
	   */
	  public void setShrinking(boolean value) {
	    m_Shrinking = value;
	  }
	  
	  /**
	   * whether to use the shrinking heuristics
	   * 
	   * @return            true, if shrinking is used
	   */
	  public boolean getShrinking() {
	    return m_Shrinking;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String shrinkingTipText() {
	    return "Whether to use the shrinking heuristic.";
	  }
	  
	  /**
	   * whether to normalize input data
	   * 
	   * @param value       whether to normalize the data
	   */
	  public void setNormalize(boolean value) {
	    m_Normalize = value;
	  }
	  
	  /**
	   * whether to normalize input data
	   * 
	   * @return            true, if the data is normalized
	   */
	  public boolean getNormalize() {
	    return m_Normalize;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String normalizeTipText() {
	    return "Whether to normalize the data.";
	  }
	  
	  /**
	   * Sets the parameters C of class i to weight[i]*C, for C-SVC (default 1).
	   * Blank separated list of doubles.
	   * 
	   * @param weightsStr          the weights (doubles, separated by blanks)
	   */
	  public void setWeights(String weightsStr) {
	    StringTokenizer       tok;
	    int                   i;
	    
	    tok           = new StringTokenizer(weightsStr, " ");
	    m_Weight      = new double[tok.countTokens()];
	    m_WeightLabel = new int[tok.countTokens()];
	    
	    if (m_Weight.length == 0)
	      System.out.println(
	          "Zero Weights processed. Default weights will be used");
	    
	    for (i = 0; i < m_Weight.length; i++) {
	      m_Weight[i]      = Double.parseDouble(tok.nextToken());
	      m_WeightLabel[i] = i;
	    }
	  }
	  
	  /**
	   * Gets the parameters C of class i to weight[i]*C, for C-SVC (default 1).
	   * Blank separated doubles.
	   * 
	   * @return            the weights (doubles separated by blanks)
	   */
	  public String getWeights() {
	    String      result;
	    int         i;
	    
	    result = "";
	    for (i = 0; i < m_Weight.length; i++) {
	      if (i > 0)
	        result += " ";
	      result += Double.toString(m_Weight[i]);
	    }
	    
	    return result;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String weightsTipText() {
	    return "The weights to use for the classes, if empty 1 is used by default.";
	  }
	  
	  /**
	   * Returns whether probability estimates are generated instead of -1/+1 for 
	   * classification problems.
	   * 
	   * @param value       whether to predict probabilities
	   */
	  public void setProbabilityEstimates(boolean value) {
	    m_ProbabilityEstimates = value;
	  }
	  
	  /**
	   * Sets whether to generate probability estimates instead of -1/+1 for 
	   * classification problems.
	   * 
	   * @return            true, if probability estimates should be returned
	   */
	  public boolean getProbabilityEstimates() {
	    return m_ProbabilityEstimates;
	  }
	  
	  /**
	   * Returns the tip text for this property
	   *
	   * @return tip text for this property suitable for
	   *         displaying in the explorer/experimenter gui
	   */
	  public String probabilityEstimatesTipText() {
	    return "Whether to generate probability estimates instead of -1/+1 for classification problems.";
	  }
	  
	
	
	public double[] distributionForInstance(Instance inst) throws Exception {
	  double tmp_distributionForInstance[]= new double[inst.numClasses()]; 
	  for (int i_att=0;(i_att<inst.numAttributes()-1);i_att++){
	     if (!use_this_split[i_att])
	                        continue;
	     
	     if (svm_structure[i_att]==1)
	       tmp_distributionForInstance= additiondouble(tmp_distributionForInstance,svm1[i_att].distributionForInstance(inst),inst.numClasses());
	     else if (svm_structure[i_att]==2)
	       tmp_distributionForInstance= additiondouble(tmp_distributionForInstance,svm2[i_att].distributionForInstance(inst),inst.numClasses());
	     else {
	    	int subset = modSplit[i_att].whichSubset(inst);
	  	if (subset ==0)
	  	  tmp_distributionForInstance= additiondouble(tmp_distributionForInstance,svm1[i_att].distributionForInstance(inst),inst.numClasses());
		else 
		  tmp_distributionForInstance= additiondouble(tmp_distributionForInstance,svm2[i_att].distributionForInstance(inst),inst.numClasses());
	  	}  
	  }
	  for (int i=0;i<inst.numClasses();i++)
	    tmp_distributionForInstance[i]=tmp_distributionForInstance[i]/pick_topn;
	  
	  return tmp_distributionForInstance;
	}
	
	private double[] additiondouble(double x1[], double x2[],int x_length){
	  double result[] = new double[x_length];
	  for (int i=0;i<x_length;i++)
	    result[i]=x1[i]+x2[i];
	  
	  return result;
	}
	public String toString() {
	    return "SubspaceSVM develpoed by Tzu-Cheng Chuang";
	  }
	/**
	   * Main method for testing this class.
	   */
	public static void main(String[] argv) {
	    runClassifier(new SubspaceSVM(), argv);
	}
	
	
	/**
	   * Returns the revision string.
	   * 
	   * @return		the revision
	   */
	public String getRevision() {
	    return RevisionUtils.extract("$Revision: 1.0 $");
	}

}


  
  
