/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    uniSplit.java
 *    Copyright (C) 2008
 *
 */

package weka.classifiers.TzuCheng.tpst;

import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;
import weka.core.matrix.Matrix;

import java.util.Enumeration;

/**
 * Class implementing a tpst.
 *
 * @author Tzu-Cheng Chuang
 * @version $Revision: 1.0 $
 */
public class MultiSplit extends ClassifierSplitModel{

  /** for serialization */
  private static final long serialVersionUID = -1278743919563022474L;

    
  /** Value of split point. */
  private double m_splitPoint=0;  
  
  /** Attribute to split on. */
  private int m_attIndex;        

  /** Minimum number of objects in a split.   */ 
  private int m_minNoObj;         

  /** Values of the coefficients to be multiplied by each attribute */
  private double m_coef[];  
  
  /** Which classes are grouped into one group */
  private int m_class_in_group[];  
  private int m_class_in_group2[];
  
  /** InfoGain of split. */
  private double m_infoGain; 

  /** GainRatio of split.  */
  private double m_gainRatio;  

  /** The sum of the weights of the instances. */
  private double m_sumOfWeights;  

  /** Static reference to splitting criterion. */
  private static InfoGainSplitCrit m_infoGainCrit = new InfoGainSplitCrit();

  /** Static reference to splitting criterion. */
  private static GainRatioSplitCrit m_gainRatioCrit = new GainRatioSplitCrit();

  /**
   * Initializes the split model.
   */
  
  public MultiSplit(int minNoObj,double sumOfWeights){

        
    // Set minimum number of objects.
    m_minNoObj = minNoObj;

    // Set sum of weights;
    m_sumOfWeights = sumOfWeights;
  }

  
  public void buildClassifier(Instances trainInstances,int[] class_in_group){
    try{
      m_class_in_group= class_in_group;
      m_class_in_group2= new int[trainInstances.numClasses()-class_in_group.length];
      m_coef=new double[trainInstances.numAttributes()];
      
      int k=0;
      for (int i=0;i<trainInstances.numClasses();i++){
	boolean in_the_group=false;
	//check whether i is in the group
	for (int j=0;j<class_in_group.length;j++){
	  if (i==class_in_group[j]){
	    in_the_group=true;
	  }
	}
	if (!in_the_group){
	  m_class_in_group2[k++]=i;
      
	}
      }
      buildClassifier(trainInstances);
    }
    catch(Exception e){
      e.printStackTrace();
    }
  }
  /**
   * Creates a C4.5-type split on the given data.
   *
   * @exception Exception if something goes wrong
   */
  public void buildClassifier(Instances trainInstances)
       throws Exception {

    // Initialize the remaining instance variables.
    m_numSubsets = 0;
    m_splitPoint = Double.MAX_VALUE;
    m_infoGain = 0;
    m_gainRatio = 0;

    // Different treatment for enumerated and numeric attributes.
    // Usually multisplit doesn't allow nominal attributes
    if (trainInstances.attribute(m_attIndex).isNominal()){
      handleEnumeratedAttribute(trainInstances);
    }else{
      //trainInstances.sort(trainInstances.attribute(m_attIndex));
      handleNumericAttribute(trainInstances);
    }
  }    

  /**
   * Returns index of attribute for which split was generated.
   */
  public final int attIndex(){

    return m_attIndex;
  }
  
  /**
   * Returns (C4.5-type) gain ratio for the generated split.
   */
  public final double gainRatio(){
    return m_gainRatio;
  }

  /**
   * Gets class probability for instance.
   *
   * @exception Exception if something goes wrong
   */
  public final double classProb(int classIndex,Instance instance,
				int theSubset) throws Exception {

    if (theSubset <= -1) {
      double [] weights = weights(instance);
      if (weights == null) {
	return m_distribution.prob(classIndex);
      } else {
	double prob = 0;
	for (int i = 0; i < weights.length; i++) {
	  prob += weights[i] * m_distribution.prob(classIndex, i);
	}
	return prob;
      }
    } else {
      if (Utils.gr(m_distribution.perBag(theSubset), 0)) {
	return m_distribution.prob(classIndex, theSubset);
      } else {
	return m_distribution.prob(classIndex);
      }
    }
  }
 
  /**
   * Creates split on enumerated attribute.
   *
   * @exception Exception if something goes wrong
   */
  private void handleEnumeratedAttribute(Instances trainInstances)
       throws Exception {
    
    Distribution newDistribution,secondDistribution;
    int numAttValues;
    double currIG,currGR;
    Instance instance;
    int i;

    numAttValues = trainInstances.attribute(m_attIndex).numValues();
    newDistribution = new Distribution(numAttValues,
				       trainInstances.numClasses());
    
    // Only Instances with known values are relevant.
    Enumeration enu = trainInstances.enumerateInstances();
    while (enu.hasMoreElements()) {
      instance = (Instance) enu.nextElement();
      if (!instance.isMissing(m_attIndex))
	newDistribution.add((int)instance.value(m_attIndex),instance);
    }
    m_distribution = newDistribution;

    // For all values
    for (i = 0; i < numAttValues; i++){

      if (Utils.grOrEq(newDistribution.perBag(i),m_minNoObj)){
	secondDistribution = new Distribution(newDistribution,i);
	
	// Check if minimum number of Instances in the two
	// subsets.
	if (secondDistribution.check(m_minNoObj)){
	  m_numSubsets = 2;
	  currIG = m_infoGainCrit.splitCritValue(secondDistribution,
					       m_sumOfWeights);
	  currGR = m_gainRatioCrit.splitCritValue(secondDistribution,
						m_sumOfWeights,
						currIG);
	  if ((i == 0) || Utils.gr(currGR,m_gainRatio)){
	    m_gainRatio = currGR;
	    m_infoGain = currIG;
	    m_splitPoint = (double)i;
	    m_distribution = secondDistribution;
	  }
	}
      }
    }
  }
 
  
  public int[] getGroups(){
    return m_class_in_group;
  }
  
  
  /**
   * Creates split on numeric attribute.
   *
   * @exception Exception if something goes wrong
   */
  private void handleNumericAttribute(Instances trainInstances)
       throws Exception {
  
    int next = 1;
    int last = 0;
    int index = 0;
    int splitIndex = -1;
    double currentInfoGain;
    double defaultEnt;
    double minSplit;
  
    m_distribution =new Distribution(trainInstances);
      

    // Compute minimum number of Instances required in each
    // subset.
    minSplit =  0.1*(m_distribution.total())/((double)trainInstances.numClasses());
    if (Utils.smOrEq(minSplit,m_minNoObj)){ 
      minSplit = m_minNoObj;
    }
    else{
      if (Utils.gr(minSplit,25)) 
	minSplit = 25;
    }
    
    
    //Find the best coefficients by using linear discriminant analysis
    InstancesMatrix inst_matrix= new InstancesMatrix(trainInstances);
    
    Matrix mOnlyData=inst_matrix.getMatrix_onlyData ();
    
    Matrix matG1=inst_matrix.getMatrix_InClass (m_class_in_group);
    Matrix matG2=inst_matrix.getMatrix_InClass (m_class_in_group2);
    
    MatrixExt mC= new MatrixExt(matG1);
    Matrix mG1_mean= mC.mean(matG1);
    Matrix mG2_mean= mC.mean(matG2);
    Matrix mG1_mean_minus_mG2_mean=mG1_mean.minus(mG2_mean);
    
    Matrix withinScatter= mC.cov(matG1).plus(mC.cov(matG2));
    
    Matrix w_coef=mG1_mean_minus_mG2_mean.times(mC.pinv(withinScatter).transpose());
    
    Matrix w_0=mG1_mean.plus(mG2_mean).times(-0.5).times(w_coef.transpose());
    
    Matrix transformed =mOnlyData.times(w_coef.transpose()).plus(mC.repmat(w_0, mOnlyData.getRowDimension(), 1));
    
    for (int jj=0;jj<trainInstances.numAttributes()-1;jj++){
      m_coef[jj]=w_coef.get(0,jj);
    }
    m_coef[trainInstances.numAttributes()-1]=w_0.get(0,0);
    
    
    
    
    // Compute values of criteria for all possible split
    // indices.
    defaultEnt = m_infoGainCrit.oldEnt(m_distribution);
    
    // Construct the distribution based on found coefficients
    Distribution local_distribution = new Distribution(2,trainInstances.numClasses());
    for (int i_inst=0;i_inst<trainInstances.numInstances();i_inst++){
      Instance local_instance=trainInstances.instance(i_inst);
      local_distribution.add(whichSubset(local_instance), local_instance);
    }
    
    m_distribution = local_distribution;
	
	// Check if enough Instances in each subset and compute
	// values for criteria.
    /*
	if (Utils.grOrEq(m_distribution.perBag(0),minSplit) && 
	    Utils.grOrEq(m_distribution.perBag(1),minSplit)){
	  currentInfoGain = m_infoGainCrit.splitCritValue(m_distribution,m_sumOfWeights,defaultEnt);
	  if (Utils.gr(currentInfoGain,m_infoGain)){
	    m_infoGain = currentInfoGain;
	  }
	}
	*/ 
	m_infoGain = m_infoGainCrit.splitCritValue(m_distribution,m_sumOfWeights,defaultEnt);
    
   
    
    // Compute modified gain ratio for best split.
    m_gainRatio = m_gainRatioCrit.splitCritValue(m_distribution,m_sumOfWeights,m_infoGain);
    m_numSubsets = 2;
    m_splitPoint = 0;
  }

  /**
   * Returns (C4.5-type) information gain for the generated split.
   */
  public final double infoGain(){

    return m_infoGain;
  }

  /**
   * Prints left side of condition.
   * 
   * @param data the data to get the attribute name from.
   * @return the attribute name
   */
  public final String leftSide(Instances data){
    
    StringBuffer text;

    text = new StringBuffer();
    text.append("**");
    for (int i=0;i<m_coef.length;i++){
      text.append(" "+m_coef[i]);
    }
    
    return text.toString();
  }

  /**
   * Prints the condition satisfied by instances in a subset.
   *
   * @param index of subset and training set.
   */
  public final String rightSide(int index,Instances data){

    StringBuffer text;

    text = new StringBuffer();
    if (data.attribute(m_attIndex).isNominal()){
      if (index == 0)
	text.append(" = "+
		    data.attribute(m_attIndex).value((int)m_splitPoint));
      else
	text.append(" != "+
		    data.attribute(m_attIndex).value((int)m_splitPoint));
    }else
      if (index == 0)
	text.append(" <= "+m_splitPoint);
      else
	text.append(" > "+m_splitPoint);
    
    return text.toString();
  }


 
  /**
   * Returns a string containing java source code equivalent to the test
   * made at this node. The instance being tested is called "i".
   *
   * @param index index of the nominal value tested
   * @param data the data containing instance structure info
   * @return a value of type 'String'
   */
  public final String sourceExpression(int index, Instances data) {

    StringBuffer expr = null;
    if (index < 0) {
      return "i[" + m_attIndex + "] == null";
    }
    if (data.attribute(m_attIndex).isNominal()) {
      if (index == 0) {
	expr = new StringBuffer("i[");
      } else {
	expr = new StringBuffer("!i[");
      }
      expr.append(m_attIndex).append("]");
      expr.append(".equals(\"").append(data.attribute(m_attIndex)
				     .value((int)m_splitPoint)).append("\")");
    } else {
      expr = new StringBuffer("((Double) i[");
      expr.append(m_attIndex).append("])");
      if (index == 0) {
	expr.append(".doubleValue() <= ").append(m_splitPoint);
      } else {
	expr.append(".doubleValue() > ").append(m_splitPoint);
      }
    }
    return expr.toString();
  }  

  /**
   * Sets split point to greatest value in given data smaller or equal to
   * old split point.
   * (C4.5 does this for some strange reason).
   */
  public final void setSplitPoint(Instances allInstances){
    
    double newSplitPoint = -Double.MAX_VALUE;
    double tempValue;
    Instance instance;
    
    if ((!allInstances.attribute(m_attIndex).isNominal()) &&(m_numSubsets > 1)){
      Enumeration enu = allInstances.enumerateInstances();
      while (enu.hasMoreElements()) {
	instance = (Instance) enu.nextElement();
	if (!instance.isMissing(m_attIndex)){
	  tempValue = instance.value(m_attIndex);
	  if (Utils.gr(tempValue,newSplitPoint) && 
	      Utils.smOrEq(tempValue,m_splitPoint))
	    newSplitPoint = tempValue;
	}
      }
      m_splitPoint = newSplitPoint;
    }
  }
  
  /**
   * Sets distribution associated with model.
   */
  public void resetDistribution(Instances data) throws Exception {

    Instances insts = new Instances(data, data.numInstances());
    for (int i = 0; i < data.numInstances(); i++) {
      if (whichSubset(data.instance(i)) > -1) {
	insts.add(data.instance(i));
      }
    }
    Distribution newD = new Distribution(insts, this);
    newD.addInstWithUnknown(data, m_attIndex);
    m_distribution = newD;
  }

  /**
   * Returns weights if instance is assigned to more than one subset.
   * Returns null if instance is only assigned to one subset.
   */
  public final double [] weights(Instance instance){
    
    double [] weights;
    int i;
    
    if (instance.isMissing(m_attIndex)){
      weights = new double [m_numSubsets];
      for (i=0;i<m_numSubsets;i++)
	weights [i] = m_distribution.perBag(i)/m_distribution.total();
      return weights;
    }else{
      return null;
    }
  }
  
  /**
   * Returns 0 is it belongs to left side (Fisher value is less than or equal to 0)
   * Otherwise, returns 1
   * Returns -1 if instance is assigned to more than one subset.
   *
   * @exception Exception if something goes wrong
   */

  public final int whichSubset(Instance instance) throws Exception {
    
	if (Utils.smOrEq(fisher_value(instance),0))
	  return 0;
	else
	  return 1;

  }
  public final double fisher_value(Instance instance) {
    double []instanceData=new double[instance.numAttributes()-1];
    int k=0;
    for (int i=0;i<instance.numAttributes();i++){
    	if(i!=instance.classIndex()){
    	  instanceData[k++]=instance.value(i);
    	}
    }
    double ans=0;
    for(int j=0;j<instanceData.length;j++){
      ans+=instanceData[j]*m_coef[j];
    }
    ans+=m_coef[m_coef.length-1];
    return ans;
  }
}
